#!/usr/bin/env python3

import yaml
import sys
from pathlib import Path
from collections import defaultdict
from math import ceil
import numpy as np


def ingest(folder, pattern='*.yaml'):
    p = Path(folder)
    data = defaultdict(lambda: defaultdict(dict))
    files = []
    for subfolder in p.iterdir():
        if not subfolder.is_dir():
            continue
        files += sorted(subfolder.glob(pattern))
    for i, f in enumerate(files):
        print("     [%d/%d] %s" % (i + 1, len(files), f), flush=True)
        loaded = yaml.safe_load(f.open())
        if 'task set' in loaded:
            tid = loaded['task set id']
            data[tid]['task set'] = loaded['task set']
        elif 'results' in loaded:
            tid = loaded['task set id']
            data[tid]['utilization'] = loaded['utilization']
            pm = loaded['preemption model']
            sched = loaded['scheduling policy']
            data[tid][pm][sched] = loaded['results']
        else:
            assert False # unexpected YAML file
    return data

def check_all_present(data):
    for tsid in data.keys():
        workload = data[tsid]
        if not 'utilization' in workload:
            print("Workload %s is missing results." % tsid)
            return False
        if not 'task set' in workload:
            print("Workload %s is incomplete." % tsid)
            return False
    return True

def tcount(workload):
    return len(workload['task set'])

def util(workload):
    return workload['utilization']

def feasible(workload):
    return util(workload) <= 1.0

def fifo_bounded(workload):
    return all(type(t['response-time bound']) == int for t in workload['NP']['FIFO'])

def afifo_bounded(workload):
    return all(type(t['response-time bound']) == int for t in workload['NP']['AFIFO'])

def infer_rate(t):
    return t['period'] if 'period' in t else (t['deadline'] if 'deadline' in t else t['minimum inter-arrival time'])

def aggregate_by_tcount(data, predicates):
    bins = defaultdict(lambda: defaultdict(list))
    for tsid in data:
        workload = data[tsid]
        n = tcount(workload)
        for label, pred in predicates:
            bins[n][label].append(pred(workload))
    rows = [[n] + [np.mean(bins[n][label]) for label, _ in predicates]
            for n in sorted(bins.keys())]
    header = ["TCOUNT"] + [label for label, _ in predicates]
    return (rows, header)

def aggregate_by_util(data, predicates, granularity=0.025):
    bins = defaultdict(lambda: defaultdict(list))
    for tsid in data:
        workload = data[tsid]
        u = util(workload)
        b = ceil(u / granularity)
        for label, pred in predicates:
            bins[b][label].append(pred(workload))
    rows = [[b * granularity] + [np.mean(bins[b][label]) for label, _ in predicates]
            for b in sorted(bins.keys())]
    header = ["UTIL"] + [label for label, _ in predicates]
    return (rows, header)

def write_csv(rows, header=None, f=sys.stdout):
    if header:
        print(", ".join(("%10s" % h for h in header)), file=f)
    for row in rows:
        print(", ".join(("%10.3f" % d for d in row)), file=f)


AFIFO_COMPARISON = [
    ('FEAS', feasible),
    ('FIFO', fifo_bounded),
    ('AFIFO', afifo_bounded),
]

def aggregate_bound_stats_by_tcount(data, baseline_policy, eval_policy, stats):
    bpm, bs = baseline_policy
    epm, es = eval_policy
    bins = defaultdict(lambda: defaultdict(list))
    all_rates = set()
    for tsid in data:
        workload = data[tsid]
        if not fifo_bounded(workload):
            continue
        n = tcount(workload)
        for (t, res_b, res_e) in zip(workload['task set'], workload[bpm][bs], workload[epm][es]):
            assert t['id'] == res_b['task id'] == res_e['task id']
            rate = infer_rate(t)
            # both unbounded, or both bounded
            assert type(res_e['response-time bound']) == type(res_b['response-time bound'])
            if type(res_e['response-time bound']) != int:
                continue
            ratio = res_e['response-time bound'] / res_b['response-time bound']
            bins[n][rate].append(ratio)
            all_rates.add(rate)
    all_rates = sorted(all_rates)
    rows = [[n] + [stat(bins[n][rate])
                   for rate in all_rates
                   for _, stat in stats]
            for n in sorted(bins.keys())]
    header = ["TCOUNT"] + ["%s-%d" % (label, int(rate / 10**6))
                           for rate in all_rates for label, _ in stats]
    return (rows, header)

def aggregate_bound_stats_by_util(data, baseline_policy, eval_policy, stats, granularity=0.025):
    bpm, bs = baseline_policy
    epm, es = eval_policy
    bins = defaultdict(lambda: defaultdict(list))
    all_rates = set()
    for tsid in data:
        workload = data[tsid]
        if not fifo_bounded(workload):
            continue
        u = util(workload)
        b = ceil(u / granularity)
        for (t, res_b, res_e) in zip(workload['task set'], workload[bpm][bs], workload[epm][es]):
            assert t['id'] == res_b['task id'] == res_e['task id']
            rate = infer_rate(t)
            # both unbounded, or both bounded
            assert type(res_e['response-time bound']) == type(res_b['response-time bound'])
            if type(res_e['response-time bound']) != int:
                continue
            ratio = res_e['response-time bound'] / res_b['response-time bound']
            bins[b][rate].append(ratio)
            all_rates.add(rate)
    all_rates = sorted(all_rates)
    rows = [[b * granularity] + [stat(bins[b][rate])
                   for rate in all_rates
                   for _, stat in stats]
            for b in sorted(bins.keys())]
    header = ["UTIL"] + ["%s-%d" % (label, int(rate / 10**6))
                           for rate in all_rates for label, _ in stats]
    return (rows, header)

RT_COMPARISON_STATS = [
    ('CNT', len),
    ('MIN', lambda a: np.min(a) if a else -1),
    ('AVG', lambda a: np.mean(a) if a else -1),
    ('MED', lambda a: np.median(a) if a else -1),
    ('P95', lambda a: np.percentile(a, 95) if a else -1),
    ('P99', lambda a: np.percentile(a, 99) if a else -1),
    ('MAX', lambda a: np.max(a) if a else -1),
    ('STD', lambda a: np.std(a) if a else -1),
]

def eval_fifo_comparison(folder, pattern='*.yaml'):
    print("[II] Ingesting %s/%s... " % (folder, pattern))
    data = ingest(folder, pattern)

    if not check_all_present(data):
        print("[EE] Aborting %s/%s. " % (folder, pattern))
        sys.exit(1)

    file = Path(folder, 'ratios_by_tcount.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_by_tcount(data, AFIFO_COMPARISON)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'ratios_by_util.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_by_util(data, AFIFO_COMPARISON)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_fp_by_tcount.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'FP'), ('NP', 'FIFO'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_fp_by_util.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_util(data, ('NP', 'FP'), ('NP', 'FIFO'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_edf_by_tcount.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'EDF'), ('NP', 'FIFO'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_edf_by_util.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_util(data, ('NP', 'EDF'), ('NP', 'FIFO'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_edf-vs-fp_by_tcount.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'FP'), ('NP', 'EDF'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    file = Path(folder, 'rtr_edf-vs-fp_by_util.csv')
    print("[II] Aggregating results in %s..." % file, flush=True)
    rows, header = aggregate_bound_stats_by_util(data, ('NP', 'FP'), ('NP', 'EDF'), RT_COMPARISON_STATS)
    write_csv(rows, header, f=file.open('w'))

    for label, stat in RT_COMPARISON_STATS:
        file = Path(folder, 'rtr_fp_%s_by_tcount.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'FP'), ('NP', 'FIFO'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

        file = Path(folder, 'rtr_fp_%s_by_util.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_util(data, ('NP', 'FP'), ('NP', 'FIFO'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

        file = Path(folder, 'rtr_edf_%s_by_tcount.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'EDF'), ('NP', 'FIFO'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

        file = Path(folder, 'rtr_edf_%s_by_util.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_util(data, ('NP', 'EDF'), ('NP', 'FIFO'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

        file = Path(folder, 'rtr_edf-vs-fp_%s_by_tcount.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_tcount(data, ('NP', 'FP'), ('NP', 'EDF'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

        file = Path(folder, 'rtr_edf-vs-fp_%s_by_util.csv' % label)
        print("[II] Aggregating results in %s..." % file, flush=True)
        rows, header = aggregate_bound_stats_by_util(data, ('NP', 'FP'), ('NP', 'EDF'), [(label, stat)])
        write_csv(rows, header, f=file.open('w'))

    print("[II] Done with %s/%s. " % (folder, pattern))

if __name__ == "__main__":
    eval_fifo_comparison('workloads/ex1', pattern='*.yaml')
    eval_fifo_comparison('workloads/ex2', pattern='*.yaml')
    eval_fifo_comparison('workloads/ex3', pattern='*.yaml')
    eval_fifo_comparison('workloads/ex4', pattern='*.yaml')
