#!/usr/bin/env python3

from pathlib import Path

import numpy as np
import pandas as pd

import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.ticker import (FormatStrFormatter, FixedLocator)


figsize = (8,6)

results_path = Path.cwd().joinpath('results')
if not results_path.exists(): raise RuntimeError('No directory with results')
plots_path = Path.cwd().joinpath('plots')
if not plots_path.exists(): plots_path.mkdir()

palette_long = sns.color_palette("husl", 9)
palette_short = sns.color_palette("husl", 5)

plt.rcParams.update({'font.size': 14})

df = []

def parse_dataframe(name):
    df = pd.read_csv(str(results_path) + name)
    df.columns = list(map(lambda idx: idx.strip(), df.columns))
    df = df.rename(columns={
        "TCOUNT": "Number of tasks", 
        "UTIL": "Total utilization", 
        "FEAS": "Feasibility", 
        "FIFO": "This paper (Theorem 1)", 
        "AFIFO": "Altmeyer et al. (2016)"})

    if 'ratios_by_tcount' in name:
        df = df.melt('Number of tasks', var_name='Analysis', value_name='Fraction of task sets')
    elif 'ratios_by_util' in name:
        df = df.melt('Total utilization', var_name='Analysis', value_name='Fraction of task sets')
    elif 'rtr_edf_AVG_by_tcount' in name:
        df = df.melt('Number of tasks', var_name='Task period', value_name='Ratio: FIFO / EDF')
        df['Ratio: EDF / FIFO'] = 1 / df['Ratio: FIFO / EDF'] 
        df['Task period'] = df['Task period'].apply(lambda label : label[4:] + ' ms')
    elif 'rtr_edf_AVG_by_util' in name:
        df = df.melt('Total utilization', var_name='Task period', value_name='Ratio: FIFO / EDF')
        df['Ratio: EDF / FIFO'] = 1 / df['Ratio: FIFO / EDF'] 
        df['Task period'] = df['Task period'].apply(lambda label : label[4:] + ' ms')
    elif 'rtr_fp_AVG_by_tcount' in name:
        df = df.melt('Number of tasks', var_name='Task period', value_name='Ratio: FIFO / FP')
        df['Ratio: FP / FIFO'] = 1 / df['Ratio: FIFO / FP'] 
        df['Task period'] = df['Task period'].apply(lambda label : label[4:] + ' ms')
    elif 'rtr_fp_AVG_by_util.csv' in name:
        df = df.melt('Total utilization', var_name='Task period', value_name='Ratio: FIFO / FP')
        df['Ratio: FP / FIFO'] = 1 / df['Ratio: FIFO / FP'] 
        df['Task period'] = df['Task period'].apply(lambda label : label[4:] + ' ms')

    return df


################################################################
#                                                              #
#                      Full period range                       #
#                                                              #
################################################################

f, (ax11) = plt.subplots(1, 1, figsize=figsize) 

del df; df = parse_dataframe('/ex1_ratios_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax11, 
    x="Number of tasks", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax11.grid(linestyle='-', color='0.95')
ax11.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_ratios_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax12) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex1_ratios_by_util.csv')

sns.lineplot(
    data=df[df['Total utilization'] <= 1.0], 
    ax=ax12, 
    x="Total utilization", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax12.grid(linestyle='-', color='0.95')
ax12.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_ratios_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax21) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex1_rtr_edf_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax21, 
    x="Number of tasks", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax21.legend(title="Task period", ncol=3)
ax21.set_yscale('log')
ax21.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax21.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax21.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax21.grid(which='both', linestyle='-', color='0.95')
ax21.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_rtr_edf_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax22) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex1_rtr_edf_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax22, 
    x="Total utilization", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax22.legend(title="Task period", ncol=3)
ax22.set_yscale('log')
ax22.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax22.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax22.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax22.grid(which='both', linestyle='-', color='0.95')
ax22.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_rtr_edf_AVG_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax31) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex1_rtr_fp_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax31, 
    x="Number of tasks", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax31.legend(title="Task period", ncol=3)
ax31.set_yscale('log')
ax31.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax31.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax31.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax31.grid(which='both', linestyle='-', color='0.95')
ax31.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_rtr_fp_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax32) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex1_rtr_fp_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax32, 
    x="Total utilization", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax32.legend(title="Task period", ncol=3)
ax32.set_yscale('log')
ax32.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax32.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax32.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax32.grid(which='both', linestyle='-', color='0.95')
ax32.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex1_rtr_fp_AVG_by_util.pdf'))
plt.close()


################################################################
#                                                              #
#             Restricted period range (Ti ≥ 20 ms)             #
#                                                              #
################################################################

f, (ax11) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex2_ratios_by_tcount.csv')

ax11.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df, 
    ax=ax11, 
    x="Number of tasks", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax11.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_ratios_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax12) = plt.subplots(1, 1, figsize=figsize) 

del df; df = parse_dataframe('/ex2_ratios_by_util.csv')

ax12.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df[df['Total utilization'] <= 1.0], 
    ax=ax12, 
    x="Total utilization", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax12.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_ratios_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax21) = plt.subplots(1, 1, figsize=figsize) 

del df; df = parse_dataframe('/ex2_rtr_edf_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax21, 
    x="Number of tasks", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax21.legend(title="Task period", ncol=3)
ax21.set_yscale('log')
ax21.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax21.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax21.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax21.grid(which='both', linestyle='-', color='0.95')
ax21.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_rtr_edf_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax22) = plt.subplots(1, 1, figsize=figsize) 

del df; df = parse_dataframe('/ex2_rtr_edf_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax22, 
    x="Total utilization", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax22.legend(title="Task period", ncol=3)
ax22.set_yscale('log')
ax22.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax22.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax22.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax22.grid(which='both', linestyle='-', color='0.95')
ax22.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_rtr_edf_AVG_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax31) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex2_rtr_fp_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax31, 
    x="Number of tasks", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax31.legend(title="Task period", ncol=3)
ax31.set_yscale('log')
ax31.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax31.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax31.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax31.grid(which='both', linestyle='-', color='0.95')
ax31.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_rtr_fp_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax32) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex2_rtr_fp_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax32, 
    x="Total utilization", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax32.legend(title="Task period", ncol=3)
ax32.set_yscale('log')
ax32.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 
                                            1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0]))
ax32.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax32.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax32.grid(which='both', linestyle='-', color='0.95')
ax32.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex2_rtr_fp_AVG_by_util.pdf'))
plt.close()


################################################################
#                                                              #
#            With release jitter JITi ∈ [0, 0.75Ti]            #
#                                                              #
################################################################
f, (ax11) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex3_ratios_by_tcount.csv')

ax11.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df, 
    ax=ax11, 
    x="Number of tasks", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax11.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_ratios_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax12) = plt.subplots(1, 1, figsize=figsize) 

del df; df = parse_dataframe('/ex3_ratios_by_util.csv')

ax12.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df[df['Total utilization'] <= 1.0], 
    ax=ax12, 
    x="Total utilization", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax12.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_ratios_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax21) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex3_rtr_edf_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax21, 
    x="Number of tasks", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax21.legend(title="Task period", ncol=3)
ax21.set_yscale('log')
ax21.yaxis.set_minor_locator(FixedLocator([0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 8.0, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]))
ax21.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax21.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax21.grid(which='both', linestyle='-', color='0.95')
ax21.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_rtr_edf_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax22) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex3_rtr_edf_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax22, 
    x="Total utilization", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax22.legend(title="Task period", ncol=3)
ax22.set_yscale('log')
ax22.yaxis.set_minor_locator(FixedLocator([0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 8.0, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]))
ax22.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax22.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax22.grid(which='both', linestyle='-', color='0.95')
ax22.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_rtr_edf_AVG_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax31) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex3_rtr_fp_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax31, 
    x="Number of tasks", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax31.legend(title="Task period", ncol=3)
ax31.set_yscale('log')
ax31.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 10.0, 20.0, 30.0, 40.0, 50.0]))
ax31.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax31.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax31.grid(which='both', linestyle='-', color='0.95')
ax31.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_rtr_fp_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax32) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex3_rtr_fp_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax32, 
    x="Total utilization", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_short,
)
ax32.legend(title="Task period", ncol=3)
ax32.set_yscale('log')
ax32.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 10.0, 20.0, 30.0, 40.0, 50.0]))
ax32.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax32.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax32.grid(which='both', linestyle='-', color='0.95')
ax32.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex3_rtr_fp_AVG_by_util.pdf'))
plt.close()


################################################################
#                                                              #
#            With release jitter JITi ∈ [0, 1.5Ti]             #
#                                                              #
################################################################
f, (ax11) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_ratios_by_tcount.csv')

ax11.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df, 
    ax=ax11, 
    x="Number of tasks", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax11.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_ratios_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax12) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_ratios_by_util.csv')

ax12.grid(linestyle='-', color='0.95')
sns.lineplot(
    data=df[df['Total utilization'] <= 1.0], 
    ax=ax12, 
    x="Total utilization", 
    y="Fraction of task sets",
    hue="Analysis",
)
ax12.lines[0].set_marker('o')

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_ratios_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax21) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_rtr_edf_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax21, 
    x="Number of tasks", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax21.legend(title="Task period", ncol=3)
ax21.set_yscale('log')
ax21.yaxis.set_minor_locator(FixedLocator([0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 8.0, 10.0, 20.0, 30.0, 40.0, 60.0, 70.0, 80.0, 90.0, 100.0]))
ax21.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax21.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax21.grid(which='both', linestyle='-', color='0.95')
ax21.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_rtr_edf_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax22) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_rtr_edf_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax22, 
    x="Total utilization", 
    y="Ratio: EDF / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax22.legend(title="Task period", ncol=3)
ax22.set_yscale('log')
ax22.yaxis.set_minor_locator(FixedLocator([0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 8.0, 10.0, 20.0, 30.0, 40.0, 60.0, 70.0, 80.0, 90.0, 100.0]))
ax22.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax22.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax22.grid(which='both', linestyle='-', color='0.95')
ax22.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_rtr_edf_AVG_by_util.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax31) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_rtr_fp_AVG_by_tcount.csv')

sns.lineplot(
    data=df, 
    ax=ax31, 
    x="Number of tasks", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax31.legend(title="Task period", ncol=3)
ax31.set_yscale('log')
ax31.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 10.0, 20.0, 30.0, 50.0, 70.0]))
ax31.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax31.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax31.grid(which='both', linestyle='-', color='0.95')
ax31.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_rtr_fp_AVG_by_tcount.pdf'))
plt.close()

# ---------------------------------------------------------------- #
f, (ax32) = plt.subplots(1, 1, figsize=figsize) 
del df; df = parse_dataframe('/ex4_rtr_fp_AVG_by_util.csv')

sns.lineplot(
    data=df, 
    ax=ax32, 
    x="Total utilization", 
    y="Ratio: FP / FIFO",
    hue="Task period",
    palette=palette_long,
)
ax32.legend(title="Task period", ncol=3)
ax32.set_yscale('log')
ax32.yaxis.set_minor_locator(FixedLocator([0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1.0, 2.0, 3.0, 4.0, 5.0, 7.0, 10.0, 20.0, 30.0, 50.0, 70.0]))
ax32.yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
ax32.yaxis.set_minor_formatter(FormatStrFormatter('%.1f'))
ax32.grid(which='both', linestyle='-', color='0.95')
ax32.axhline(y=1, color='r', linestyle='--', linewidth=1, alpha=0.5)

plt.tight_layout()
plt.savefig(plots_path.joinpath('ex4_rtr_fp_AVG_by_util.pdf'))
plt.close()
