# RTA Tool

A command-line tool to perform a response-time analysis (RTA) on a task set given in a simple YAML dialect.

The actual analysis is implemented in [the `response-time-analysis` crate](https://crates.io/crates/response-time-analysis).

## Compilation, Documentation, Tests

To compile the tool, run:

```
cargo build --release
```

The resulting executable binary is `target/release/rta-tool`.

To render the documentation, run:

```
cargo doc
```

To open the documentation, run:

```
cargo doc --open
```

To run all tests, run:

```
cargo test
```

## Input Format

The tool operates on a simple YAML-based input format that is mostly self-explanatory. Here is an example:

```YAML
---
task set id: My Example Workload
scheduling policy: EDF
preemption model: fully preemptive
analysis horizon: 2000
task set:

- id: Task 1
  period: 100
  worst-case execution time: 24
  deadline: 100

- id: Task 2
  period: 50
  worst-case execution time: 7
  deadline: 30

- id: Task 3
  minimum inter-arrival time: 75
  release jitter: 25
  worst-case execution time: 10
  deadline: 75

- id: Task 4
  period: 500
  worst-case execution time: 100
  deadline: 1000

- id: Task 5
  arrival curve: [500, [[1, 1], [15, 2], [100, 3]]]
  worst-case execution time: 5
  deadline: 15
...
```

Some points to note:

- The task-set and task IDs can be arbitrary strings.
- The workload is assumed to be scheduled with the `EDF` policy, as indicated by the `scheduling policy` key. See below for further policies.
- The workload is assumed to comprise `fully preemptive` tasks, as indicated by the `preemption model`. See below for further models.
- Tasks 1, 2, and 4 are declared to be periodic tasks. 
- Task 3 is declared to be a sporadic task affected by jitter. The jitter parameter is optional and can be omitted, in which ase it is assumed to be zero.
- Task 5 is declared to be a sporadic task characterized by an arrival curve. Specifically, an arrival-curve prefix is given up to an interval length of at most 500 time units, with the interpretation that
	- 0 jobs are released in an interval of length 0,
	- 1 job can be released in any interval of length 1–14,
	- 2 jobs can be released in an interval of length 15–99, and.
	- 3 jobs can be released in an interval of length 100–500. 
- All numeric task parameters must be non-negative integers. For ease of explanation, the parameters here are human-readable "small" numbers, but usually they are given in nanosecond or processor-cycle resolution. 

### Supported Scheduling Policies 

The tool presently supports the following scheduling models:

1. `earliest-deadline first` or `EDF` 
2. `fixed priority` or `FP` 
3. `first-in first-out`, `first-come first-serve`, `FIFO`, or `FCFS` 
4. `AFIFO`

Policy names may be abbreviated or spelled out as indicated. 

`AFIFO` is FIFO scheduling, but using the baseline analysis proposed by Altmeyer et al. in a [tech report ](https://opus.bibliothek.uni-augsburg.de/opus4/files/69804/FIFO_scheduling_TR.pdf) in 2016.

### Supported Preemption Models

The tool presently supports the following preemption models:

1. `fully preemptive` or `FP`
2. `fully non-preemptive` or `NP`
3. `limited-preemptive` or `LP`
4. `floating non-preemptive sections` or `NPS`

All preemption models may be combined with all scheduling policies. However, note that FIFO scheduling inherently avoids preemptions, so that the choice of preemption model supported by the workload has no impact on the resulting schedule and corresponding analysis under FIFO scheduling.

## Running the Tool

To analyze a YAML file like the one given above, simply pass the name to `rta-tool`. For example, if the above snippet is stored in a file called `test.yaml`, then simply run:

```
rta-tool test.yaml
```

As is typical for Unix tools, if there is no output on the terminal, then the tool completed successfully. 

Run `rta-tool --help` to see a list of supported flags and options. 


## Output Format 

The tool stores the analysis results in a simple self-explanatory YAML format. For example, running the tool as described above stores the following results in a file named `results-test.yaml`.

```YAML
---
task set id: My Example Workload
scheduling policy: EDF
preemption model: FP
analysis horizon: 2000
utilization: 0.7433333333333334
results:
  - task id: Task 1
    response-time bound: 58
  - task id: Task 2
    response-time bound: 12
  - task id: Task 3
    response-time bound: 33
  - task id: Task 4
    response-time bound: 269
  - task id: Task 5
    response-time bound: 5
```

Some points to note:

- The workload and task IDs are simply copied from the input, to allow for easy  processing of results. 
- The analysis assumptions (scheduling policy, preemption model, analysis horizon) are noted in the results file. 
- The total utilization is recorded for convenience and further processing. 
- For each task, the computed response-time bound is noted. 

If a response-time bound cannot be found, the result would instead look something like this:

```YAML
[…]
results:
  - task id: Task 1
    response-time bound: none - no fixed point less than 2000 found for offset 0
[…]
```

