//! "glue code" connecting our [TaskSet] model with the RTAs provided by the [response-time analysis crate][rta].

use rta::arrival::ArrivalBound;
use rta::demand;
use rta::fixed_point::{SearchFailure, SearchResult};
use rta::time::{Duration, Service};

use crate::error::AppError;
use crate::io::{PreemptionModel, SchedulingPolicy};
use crate::model::{Task, TaskSet};

impl TaskSet {

    /// Run a response-time analysis for the task set assuming the given
    /// preemption model `pm` and scheduling policy `sp`. For each task,
    /// this will return a response-time bound, provided one exists
    /// below the analysis horizon specified by `limit`.
    pub fn rta(
        &self,
        pm: PreemptionModel,
        sp: SchedulingPolicy,
        limit: Duration,
    ) -> Result<Vec<SearchResult>, AppError> {
        // dispatch by preemption model and scheduling policy
        match (sp, pm) {
            (SchedulingPolicy::FIFO, _) => self.rta_fifo(limit),
            (SchedulingPolicy::AFIFO, _) => self.rta_fifo_altmeyer(),
            (SchedulingPolicy::FP, PreemptionModel::FP) => self.rta_fp_fp(limit),
            (SchedulingPolicy::FP, PreemptionModel::NP) => self.rta_np_fp(limit),
            (SchedulingPolicy::FP, PreemptionModel::LP) => self.rta_lp_fp(limit),
            (SchedulingPolicy::FP, PreemptionModel::NPS) => self.rta_fp_nps(limit),
            (SchedulingPolicy::EDF, PreemptionModel::FP) => self.rta_fp_edf(limit),
            (SchedulingPolicy::EDF, PreemptionModel::NP) => self.rta_np_edf(limit),
            (SchedulingPolicy::EDF, PreemptionModel::LP) => self.rta_lp_edf(limit),
            (SchedulingPolicy::EDF, PreemptionModel::NPS) => self.rta_edf_nps(limit),
        }
    }

    /// RTA for fully-preemptive fixed-priority scheduling
    fn rta_fp_fp(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_priorities()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let interference: Vec<_> = self.hep_tasks(tsk).map(Task::rbf).collect();
                let tau = tsk.rbf();
                rta::fixed_priority::fully_preemptive::dedicated_uniproc_rta(
                    &tau,
                    &interference[..],
                    limit,
                )
            })
            .collect())
    }

    /// RTA for fully-nonpreemptive fixed-priority scheduling
    fn rta_np_fp(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_priorities()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let interference: Vec<_> = self.hep_tasks(tsk).map(Task::rbf).collect();
                let tau = rta::fixed_priority::fully_nonpreemptive::TaskUnderAnalysis {
                    wcet: tsk.cost,
                    arrivals: &tsk.arrival_bound,
                    blocking_bound: self
                        .lp_tasks(tsk)
                        .map(|t| t.cost.wcet)
                        .max()
                        .unwrap_or_else(Service::none)
                        .saturating_sub(Service::epsilon()),
                };
                rta::fixed_priority::fully_nonpreemptive::dedicated_uniproc_rta(
                    &tau,
                    &interference[..],
                    limit,
                )
            })
            .collect())
    }

    /// RTA for limited-preemptive fixed-priority scheduling
    fn rta_lp_fp(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_priorities()?;
        self.check_last_segments()?;
        self.check_max_segments()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let interference: Vec<_> = self.hep_tasks(tsk).map(Task::rbf).collect();
                let tau = rta::fixed_priority::limited_preemptive::TaskUnderAnalysis {
                    wcet: tsk.cost,
                    arrivals: &tsk.arrival_bound,
                    last_np_segment: tsk.last_np_segment.unwrap(),
                    blocking_bound: self
                        .lp_tasks(tsk)
                        .map(|t| t.max_np_segment.unwrap())
                        .max()
                        .unwrap_or_else(Service::none)
                        .saturating_sub(Service::epsilon()),
                };
                rta::fixed_priority::limited_preemptive::dedicated_uniproc_rta(
                    &tau,
                    &interference[..],
                    limit,
                )
            })
            .collect())
    }

    /// RTA for fixed-priority scheduling with floating non-preemptive sections
    fn rta_fp_nps(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_priorities()?;
        self.check_max_segments()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let interference: Vec<_> = self.hep_tasks(tsk).map(Task::rbf).collect();
                let rbf = tsk.rbf();
                let tau = rta::fixed_priority::floating_nonpreemptive::TaskUnderAnalysis {
                    rbf: &rbf,
                    blocking_bound: self
                        .lp_tasks(tsk)
                        .map(|t| t.max_np_segment.unwrap())
                        .max()
                        .unwrap_or_else(Service::none)
                        .saturating_sub(Service::epsilon()),
                };
                rta::fixed_priority::floating_nonpreemptive::dedicated_uniproc_rta(
                    &tau,
                    &interference[..],
                    limit,
                )
            })
            .collect())
    }
    
    /// RTA for fully-preemptive EDF scheduling
    fn rta_fp_edf(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let rbf = tsk.rbf();
                let tau = rta::edf::fully_preemptive::TaskUnderAnalysis {
                    deadline: tsk.deadline,
                    rbf: &rbf,
                };
                let rbfs: Vec<_> = self.tasks_other_than(tsk).map(Task::rbf).collect();
                let other_tasks: Vec<_> = self
                    .tasks_other_than(tsk)
                    .zip(rbfs.iter())
                    .map(|(t, rbf)| rta::edf::fully_preemptive::InterferingTask {
                        deadline: t.deadline,
                        rbf,
                    })
                    .collect();
                rta::edf::fully_preemptive::dedicated_uniproc_rta(&tau, &other_tasks[..], limit)
            })
            .collect())
    }

    /// RTA for fully-nonpreemptive EDF scheduling
    fn rta_np_edf(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let tau = rta::edf::fully_nonpreemptive::TaskUnderAnalysis {
                    deadline: tsk.deadline,
                    wcet: tsk.cost,
                    arrivals: &tsk.arrival_bound,
                };
                let other_tasks: Vec<_> = self
                    .tasks_other_than(tsk)
                    .map(|t| rta::edf::fully_nonpreemptive::InterferingTask {
                        deadline: t.deadline,
                        wcet: t.cost,
                        arrivals: &t.arrival_bound,
                    })
                    .collect();
                rta::edf::fully_nonpreemptive::dedicated_uniproc_rta(&tau, &other_tasks[..], limit)
            })
            .collect())
    }

    /// RTA for limited-preemptive EDF scheduling
    fn rta_lp_edf(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_last_segments()?;
        self.check_max_segments()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let tau = rta::edf::limited_preemptive::TaskUnderAnalysis {
                    deadline: tsk.deadline,
                    wcet: tsk.cost,
                    arrivals: &tsk.arrival_bound,
                    last_np_segment: tsk.last_np_segment.unwrap(),
                };
                let rbfs: Vec<_> = self.tasks_other_than(tsk).map(Task::rbf).collect();
                let other_tasks: Vec<_> = self
                    .tasks_other_than(tsk)
                    .zip(&rbfs)
                    .map(|(t, rbf)| rta::edf::limited_preemptive::InterferingTask {
                        deadline: t.deadline,
                        rbf,
                        max_np_segment: tsk.max_np_segment.unwrap(),
                    })
                    .collect();
                rta::edf::limited_preemptive::dedicated_uniproc_rta(&tau, &other_tasks[..], limit)
            })
            .collect())
    }

    /// RTA for EDF scheduling with floating non-preemptive sections
    fn rta_edf_nps(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        self.check_max_segments()?;
        Ok(self
            .tasks
            .iter()
            .map(|tsk| {
                let rbf = tsk.rbf();
                let tau = rta::edf::floating_nonpreemptive::TaskUnderAnalysis {
                    deadline: tsk.deadline,
                    rbf: &rbf,
                };
                let rbfs: Vec<_> = self.tasks_other_than(tsk).map(Task::rbf).collect();
                let other_tasks: Vec<_> = self
                    .tasks_other_than(tsk)
                    .zip(&rbfs)
                    .map(
                        |(t, rbf)| rta::edf::floating_nonpreemptive::InterferingTask {
                            deadline: t.deadline,
                            rbf,
                            max_np_segment: t.max_np_segment.unwrap(),
                        },
                    )
                    .collect();
                rta::edf::floating_nonpreemptive::dedicated_uniproc_rta(
                    &tau,
                    &other_tasks[..],
                    limit,
                )
            })
            .collect())
    }

    /// RTA for FIFO scheduling (and any preemption model)
    fn rta_fifo(&self, limit: Duration) -> Result<Vec<SearchResult>, AppError> {
        let total_demand = demand::Aggregate::new(self.tasks.iter().map(|t| t.rbf()).collect());
        let res = rta::fifo::dedicated_uniproc_rta(&total_demand, limit);
        Ok(std::iter::repeat(res).take(self.tasks.len()).collect())
    }

    /// Legacy RTA for FIFO scheduling (and any preemption model)
    ///
    /// This is an implementation of the sporadic analysis of Altmeyer
    /// et al., "The Case for FIFO Real-Time Scheduling", 2016.
    /// <https://opus.bibliothek.uni-augsburg.de/opus4/files/69804/FIFO_scheduling_TR.pdf>
    fn rta_fifo_altmeyer(&self) -> Result<Vec<SearchResult>, AppError> {
        // Altmeyer et al.'s analysis simply sums up all WCETs. (There
        // is also a "- ε" in the write-up, but under the assumption
        // that ties are broken arbitrarily, it does not apply, and is
        // hence omitted here.) This bound is valid under the assumption
        // that no task releases another job before the bound is
        // reached, i.e., under the absence of carry-in and
        // self-interference. We test that by checking that no task
        // releases more than one job in an interval corresponding to
        // the computed bound, which limits the maximum possible busy
        // window to the computed bound. 

        // (1) sum up all WCETs
        let total_wcet: Service = self.tasks.iter().map(|t| t.cost.wcet).sum();
        let bound = Duration::from(total_wcet);

        // (2) the bound is valid only if there is no self-interference
        let self_interference_possible = self
            .tasks
            .iter()
            .any(|t| t.arrival_bound.number_arrivals(bound) > 1);

        let result = if self_interference_possible {
            Err(SearchFailure::AssumptionViolated)
        } else {
            Ok(bound)
        };
        Ok(std::iter::repeat(result).take(self.tasks.len()).collect())
    }

    /// Iterate over all tasks in the task set, skipping the provided
    /// reference task `tsk`.
    fn tasks_other_than<'a>(&'a self, tsk: &'a Task) -> impl 'a + Iterator<Item = &'a Task> {
        // remove `tsk`` from consideration
        self.tasks.iter().filter(|t| !std::ptr::eq(*t, tsk))
    }

    /// Iterate over all tasks in the task set with a fixed priority
    /// *higher than or equal to* the priority of the provided reference
    /// task `tsk`, excluding the reference task `tsk` itself.
    fn hep_tasks<'a>(&'a self, tsk: &'a Task) -> impl 'a + Iterator<Item = &'a Task> {
        // remove lower-priority tasks from consideration
        self.tasks_other_than(tsk)
            .filter(|t| t.priority.unwrap() >= tsk.priority.unwrap())
    }

    /// Iterate over all tasks in the task set with a fixed priority
    /// *lower than* to the priority of the provided reference task
    /// `tsk`.
    fn lp_tasks<'a>(&'a self, tsk: &'a Task) -> impl 'a + Iterator<Item = &'a Task> {
        // consider only lower-priority tasks
        self.tasks_other_than(tsk)
            .filter(|t| t.priority.unwrap() < tsk.priority.unwrap())
    }
}
