//! Everything that can go wrong when running the tool.

use std::process::{ExitCode, Termination};

use thiserror::Error;

/// The outcome of running the tool: either nothing to report or some
/// error occurred. 
pub type AppResult = Result<(), AppError>;

/// An enumeration of everything that can possibly go wrong when running
/// this tool.
#[derive(Error, Debug)]
pub enum AppError {
    
    /// A file operation failed with some error reported by the OS.
    #[error("error: {0}")]
    OSError(#[from] std::io::Error),

    /// The tool was given a file that is not valid YAML.
    #[error("invalid input: {0}")]
    DeserializationFailure(#[from] serde_yaml::Error),

    /// The tool was given a file that is syntactically valid YAML but
    /// semantically not a valid task model. 
    #[error("invalid model: {0}")]
    InvalidModel(String),

    /// The tool was given a task model that lacks some information
    /// required for the requested analysis.
    #[error("incomplete model: {0}")]
    IncompleteModel(String),
}

impl Termination for AppError {
    fn report(self) -> ExitCode {
        match self {
            Self::OSError(_) => ExitCode::from(1),
            Self::DeserializationFailure(_) => ExitCode::from(2),
            Self::InvalidModel(_) => ExitCode::from(3),
            Self::IncompleteModel(_) => ExitCode::from(4),
        }
    }
}

impl AppError {
    /// Report that the provided model is invalid, for the reason stated
    /// in `msg`.
    pub fn invalid(msg: String) -> Self {
        AppError::InvalidModel(msg)
    }

    /// Report that the provided model is incomplete, for the reason
    /// stated in `msg`.
    pub fn missing(msg: String) -> Self {
        AppError::IncompleteModel(msg)
    }
}
