//! Support structures for YAML-based I/O with [serde].
//!
//! This module defines a number of structs and enums to serve as a as a
//! *data-definition language* (DDL, in conjunction with [serde]). That is,
//! the structs and enums in this module are not semantically meaningful
//! by themselves. Rather, they serve as "dumb" data containers that
//! must still be validated and converted into [model][crate::model]
//! types before the parsed input can be analyzed.

use std::fs;
use std::path::Path;

use parse_display::{Display, FromStr};
use serde::{Deserialize, Serialize};

use super::error::AppError;

/// This tool assumes discrete time. 
pub type Time = u64;

/// Enumeration of supported scheduling policies and analyses.
#[derive(Display, FromStr, Serialize, Deserialize, PartialEq, Debug, Copy, Clone)]
pub enum SchedulingPolicy {
    /// Earliest-Deadline First (EDF) scheduling
    #[serde(alias = "earliest-deadline first")]
    EDF,
    /// First-In First-Out (FIFO) scheduling
    #[serde(alias = "first-in first-out")]
    #[serde(alias = "first-come first-serve")]
    #[serde(alias = "FCFS")]
    FIFO,
    /// Fixed-Priority (FP) Scheduling
    #[serde(alias = "fixed priority")]
    FP,
    /// Sporadic FIFO analysis by Altmeyer et al., 2016
    AFIFO,
}

/// Enumeration of supported preemption models. 
#[derive(Display, FromStr, Serialize, Deserialize, PartialEq, Debug, Copy, Clone)]
pub enum PreemptionModel {
    /// Fully preemptive jobs — every job can be preempted at any point
    /// in time.
    #[serde(alias = "fully preemptive")]
    FP,
    /// Fully non-preemptive jobs — jobs cannot be preempted once they
    /// have started running (i.e., run-to-completion semantics).
    #[serde(alias = "fully non-preemptive")]
    NP,
    /// Limited-preemptive jobs — jobs can be preempted only at specific
    /// preemption points between non-preemptive segments. 
    #[serde(alias = "limited-preemptive")]
    LP,
    /// Preemptive jobs with floating non-preemptive sections — jobs are
    /// generally preemptive, but may contain (short) sections in which
    /// preemptions are disabled.
    #[serde(alias = "floating non-preemptive sections")]
    NPS,
}

/// The notion of a step in step-wise defined arrival curve. 
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct Step(pub Time, pub usize);

/// DDL encoding of an arrival-curve prefix
///
/// The curve is expressed as a list of monotonic steps that define the
/// curve until the horizon, after which it must be extrapolated. 
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct ArrivalCurvePrefix {
    /// The limit until which the prefix of the curve is defined. 
    pub horizon: Time,
    /// A monotonic list of points at which the curve "steps", i.e.,
    /// increases in value.
    pub steps: Vec<Step>,
}

/// Auxiliary DDL enum used to define the expected input format for
/// arrival processes using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
#[serde(untagged)]
pub enum Arrival {
    Curve {
        #[serde(rename = "arrival curve")]
        acp: ArrivalCurvePrefix,
    },

    Period {
        #[serde(rename = "period")]
        period: u64,
    },

    Sporadic {
        #[serde(rename = "minimum inter-arrival time")]
        min_inter: u64,
        #[serde(rename = "release jitter")]
        jitter: u64,
    },
}

/// Auxiliary DDL struct used to define the input format for individual
/// tasks using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct Task {
    pub id: String,

    #[serde(flatten)]
    pub arrivals: Arrival,

    #[serde(rename = "worst-case execution time")]
    pub wcet: Time,

    #[serde(rename = "deadline")]
    pub deadline: Time,

    #[serde(rename = "priority")]
    pub priority: Option<u32>,

    #[serde(rename = "max segment")]
    pub max_nonpreemptive_segment: Option<Time>,

    #[serde(rename = "last segment")]
    pub last_nonpreemptive_segment: Option<Time>,
}

/// Auxiliary DDL struct used to define the input format for entire task
/// instances using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct ProblemInstance {
    #[serde(rename = "task set id")]
    pub id: Option<String>,

    #[serde(rename = "scheduling policy")]
    pub scheduling_policy: SchedulingPolicy,

    #[serde(rename = "preemption model")]
    pub preemption_model: PreemptionModel,

    #[serde(rename = "analysis horizon")]
    pub analysis_horizon: Option<Time>,

    #[serde(rename = "task set")]
    pub tasks: Vec<Task>,
}

impl ProblemInstance {
    /// Parse a string as a YAML representation of a problem instance set using [serde].
    pub fn from_yaml(yaml: &str) -> Result<Self, AppError> {
        Ok(serde_yaml::from_str(yaml)?)
    }

    /// Parse the contents of a file as a YAML representation of a
    /// problem instance set using [serde].
    pub fn from_file<P: AsRef<Path>>(file: P) -> Result<Self, AppError> {
        let contents = fs::read_to_string(file)?;
        Self::from_yaml(contents.as_str())
    }
}

/// Auxiliary DDL enum used to define the output format for
/// response-time bounds using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
#[serde(untagged)]
pub enum Bound {
    Bound(Time),
    NoneFound(String),
}

impl From<&rta::fixed_point::SearchResult> for Bound {
    fn from(res: &rta::fixed_point::SearchResult) -> Self {
        match res {
            Ok(value) => Bound::Bound(u64::from(*value)),
            Err(err) => Bound::NoneFound(format!("none - {}", err)),
        }
    }
}

/// Auxiliary DDL struct used to define the output format for the result
/// for an individual task using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct AnalysisResult {
    #[serde(rename = "task id")]
    pub id: String,

    #[serde(rename = "response-time bound")]
    pub bound: Bound,
}

/// Auxiliary DDL struct used to define the output format for result
/// files using [serde].
#[derive(Serialize, Deserialize, PartialEq, Debug)]
pub struct AnalysisResults {
    #[serde(rename = "task set id")]
    pub id: Option<String>,

    #[serde(rename = "scheduling policy")]
    pub scheduling_policy: SchedulingPolicy,

    #[serde(rename = "preemption model")]
    pub preemption_model: PreemptionModel,

    #[serde(rename = "analysis horizon")]
    pub analysis_horizon: Time,

    pub utilization: f64,

    #[serde(rename = "results")]
    pub bounds: Vec<AnalysisResult>,
}

impl AnalysisResults {
    /// Construct a new analysis results struct for serialization. 
    pub fn new(
        id: &Option<String>,
        util: f64,
        pm: PreemptionModel,
        sp: SchedulingPolicy,
        limit: u64,
        bounds: &mut dyn Iterator<Item = (&String, &rta::fixed_point::SearchResult)>,
    ) -> Self {
        AnalysisResults {
            id: id.clone(),
            utilization: util,
            scheduling_policy: sp,
            preemption_model: pm,
            analysis_horizon: limit,
            bounds: bounds
                .map(|(id, res)| AnalysisResult {
                    id: id.clone(),
                    bound: Bound::from(res),
                })
                .collect(),
        }
    }

    /// Encode the analysis results as a string using YAML representation. 
    pub fn to_yaml(&self) -> Result<String, AppError> {
        Ok(serde_yaml::to_string(self)?)
    }

    /// Write the analysis results to a file in YAML representation. 
    pub fn write<P: AsRef<Path>>(&self, file: P) -> Result<(), AppError> {
        let yaml = self.to_yaml()?;
        fs::write(file, yaml)?;
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    pub fn parse_period() {
        let input = "period: 123";
        let res: Result<Arrival, _> = serde_yaml::from_str(input);
        assert_eq!(res.unwrap(), Arrival::Period { period: 123 });
    }

    #[test]
    pub fn parse_sporadic() {
        let input = "---
release jitter: 123
minimum inter-arrival time: 999
";
        let res: Result<Arrival, _> = serde_yaml::from_str(input);
        assert_eq!(
            res.unwrap(),
            Arrival::Sporadic {
                min_inter: 999,
                jitter: 123
            }
        );
    }

    #[test]
    pub fn parse_ac_prefix() {
        let input = "arrival curve: [10000, [[4, 27], [5, 99], [99, 123]]]";
        let res: Result<Arrival, _> = serde_yaml::from_str(input);
        assert_eq!(
            res.unwrap(),
            Arrival::Curve {
                acp: ArrivalCurvePrefix {
                    horizon: 10000,
                    steps: vec![Step(4, 27), Step(5, 99), Step(99, 123),]
                }
            }
        );
    }

    #[test]
    pub fn parse_periodic_task() {
        let input = "---
id: 99
period: 1173029
worst-case execution time: 59729
deadline: 1578636
...
";
        let res: Result<Task, _> = serde_yaml::from_str(input);
        assert_eq!(
            res.unwrap(),
            Task {
                id: 99.to_string(),
                arrivals: Arrival::Period { period: 1173029 },
                wcet: 59729,
                deadline: 1578636,
                priority: None,
                max_nonpreemptive_segment: None,
                last_nonpreemptive_segment: None,
            }
        );
    }

    #[test]
    pub fn parse_ac_task() {
        let input = "---
id: 17
arrival curve: [11111, [[0, 0], [14, 127], [15, 199], [199, 1123]]]
worst-case execution time: 59729
deadline: 1578636
priority: 13
max segment: 99
last segment: 42
...
";
        let res: Result<Task, _> = serde_yaml::from_str(input);
        assert_eq!(
            res.unwrap(),
            Task {
                id: 17.to_string(),
                arrivals: Arrival::Curve {
                    acp: ArrivalCurvePrefix {
                        horizon: 11111,
                        steps: vec![Step(0, 0), Step(14, 127), Step(15, 199), Step(199, 1123),]
                    }
                },
                wcet: 59729,
                deadline: 1578636,
                priority: Some(13),
                max_nonpreemptive_segment: Some(99),
                last_nonpreemptive_segment: Some(42),
            }
        );
    }

    #[test]
    pub fn parse_instance() {
        let input = "---
task set id: 937723115237613368079497
scheduling policy: EDF
preemption model: FP
task set:

- id: 0
  period: 1173029
  worst-case execution time: 59729
  deadline: 1578636

- id: 1
  period: 5345492
  worst-case execution time: 330336
  deadline: 11631725

- id: 2
  minimum inter-arrival time: 10213219
  release jitter: 99999
  worst-case execution time: 265382
  deadline: 18277692

- id: 3
  period: 57242209
  worst-case execution time: 3865381
  deadline: 59575504

- id: 4
  arrival curve: [500, [[0, 0], [1, 1], [15, 2], [100, 3]]]
  worst-case execution time: 6817556
  deadline: 169499260
...
";
        let res: Result<ProblemInstance, _> = ProblemInstance::from_yaml(&input);
        let pi = res.unwrap();
        assert_eq!(pi.tasks.len(), 5);
        assert_eq!(pi.scheduling_policy, SchedulingPolicy::EDF);
        assert_eq!(pi.preemption_model, PreemptionModel::FP);
        assert_eq!(pi.id, Some(String::from("937723115237613368079497")));
        assert_eq!(pi.tasks[0].id, 0.to_string());
        assert_eq!(pi.tasks[1].wcet, 330336);
        assert_eq!(pi.tasks[2].deadline, 18277692);
        assert_eq!(
            pi.tasks[2].arrivals,
            Arrival::Sporadic {
                min_inter: 10213219,
                jitter: 99999
            }
        );
        assert_eq!(pi.tasks[3].arrivals, Arrival::Period { period: 57242209 });
        assert_eq!(
            pi.tasks[4].arrivals,
            Arrival::Curve {
                acp: ArrivalCurvePrefix {
                    horizon: 500,
                    steps: vec![Step(0, 0), Step(1, 1), Step(15, 2), Step(100, 3),]
                }
            }
        );
    }
}
