//! This crate provides the `rta-tool` application for performing
//! response-time analyses (RTAs) on workloads described in simple YAML
//! files.
//!
//! This crate is straightforward in nature and doesn't do much besides
//! basic file handling (parsing input, storing output). The actual
//! analyses are implemented in the [response-time analysis crate][rta].

mod analysis;
mod error;
mod io;
mod model;

use std::ffi::OsString;
use std::path::PathBuf;
use std::process::{ExitCode, Termination};

use rta::time::Duration;
use structopt::StructOpt;

use error::AppResult;
use io::{PreemptionModel, SchedulingPolicy};

/// The default analysis horizon: one minute (in nanoseconds).
///
/// This value is used only if an input file does not state an analysis
/// horizon and no limit is provided via the `-h` flag. 
const DEFAULT_HORIZON: u64 = 60 * 1_000_000_000;


/// The entry point of the tool. 
///
/// 1. Parse the command-line arguments. 
/// 2. Process the provided input files.
fn main() -> ExitCode {
    let opts = Options::from_args();
    process_files(opts)
}

/// Definition of the command-line arguments that this tool understands.
#[derive(Debug, StructOpt)]
#[structopt(name = "rta-tool", about = "uniprocessor response-time analysis tool")]
struct Options {
    #[structopt(parse(from_os_str))]
    /// The file(s) containing the problem instance(s) to be analyzed (in YAML format).
    input_files: Vec<PathBuf>,

    #[structopt(short = "o", long = "output", name = "FILE")]
    /// Where to write the analysis results file(s).
    output_path: Option<PathBuf>,

    #[structopt(long = "prefix", name = "PREFIX")]
    /// Prefix string to include in the name of result file(s).
    output_prefix: Option<PathBuf>,

    #[structopt(short = "s", long = "scheduling-policy", name = "POLICY")]
    /// Override the scheduling policy stated in the problem description.
    scheduling_policy: Option<SchedulingPolicy>,

    #[structopt(short = "p", long = "preemption-model", name = "MODEL")]
    /// Override the preemption policy stated in the problem description.
    preemption_model: Option<PreemptionModel>,

    #[structopt(short = "h", long = "horizon", name = "HORIZON")]
    /// Limit at which to abort a non-converging fixed-point search.
    /// Any response-time bound must exist below this limit (or will not be found).
    horizon: Option<u64>,
}

/// Run the analysis for all provided input files. 
fn process_files(opts: Options) -> ExitCode {
    for f in &opts.input_files {
        let res = process_file(f, &opts);
        if let Err(e) = res {
            eprintln!("{} - {}", f.to_string_lossy(), e);
            return e.report();
        }
    }

    ExitCode::SUCCESS
}

/// Analyze one input file.
fn process_file(f: &PathBuf, opts: &Options) -> AppResult {
    // (1) parse the YAML input
    let pi = io::ProblemInstance::from_file(f)?;
    // (2) extract the semantic task model
    let ts = model::TaskSet::try_from(&pi)?;
    // (3) determine the preemption model (may be overriden by `-p`)
    let pm = opts.preemption_model.unwrap_or(pi.preemption_model);
    // (4) determine the scheduling policy (may be overriden by `-s`)
    let sp = opts.scheduling_policy.unwrap_or(pi.scheduling_policy);
    // (5) determine the analysis horizon (may be overriden with `-h`)
    let limit = opts
        .horizon
        .or(pi.analysis_horizon)
        .unwrap_or(DEFAULT_HORIZON);
    // (6) run the analysis 
    let results = ts.rta(pm, sp, Duration::from(limit))?;
    // (7) collect the analysis results
    let output = io::AnalysisResults::new(
        &pi.id,
        ts.total_utilization(),
        pm,
        sp,
        limit,
        &mut pi.tasks.iter().map(|t| &t.id).zip(results.iter()),
    );
    // (8) write the analysis results to a YAML file
    let dest = output_filename(f, opts);
    output.write(dest)?;
    Ok(())
}

/// Figure out the desired (or default) name for the output file.
fn output_filename(input: &PathBuf, opts: &Options) -> PathBuf {
    let default_name = {
        let mut name = OsString::new();
        if let Some(prefix) = &opts.output_prefix {
            name.push(prefix);
        } else {
            name.push("results-");
        }
        name.push(input.file_name().unwrap_or_else(|| "file.yaml".as_ref()));
        name
    };
    match &opts.output_path {
        Some(p) if !p.is_dir() => {
            // already points to a named file
            p.clone()
        }
        Some(p) => {
            // points to a directory => use default file name
            let mut dest = p.clone();
            dest.push(default_name);
            dest
        }
        None => {
            // no output directory specified => use input directory
            let mut dest = input.clone();
            dest.pop();
            dest.push(default_name);
            dest
        }
    }
}
