//! The semantic workload model ([Task] and [TaskSet]) underlying the tool.

use rta::arrival::{
    ArrivalBound, ArrivalCurvePrefix, Curve, ExtrapolatingCurve, Periodic, Sporadic,
};
use rta::demand::RBF;
use rta::time::{Duration, Service};
use rta::wcet;

use super::error::AppError;
use super::io;

/// A wrapper type encapsulating the supported arrival processes.
///
/// Currently, three arrival processes are supported: 
/// 1. *periodic* arrivals,
/// 2. *sporadic* arrivals (with a minimum inter-arrival time), and 
/// 3. *arbitrary* arrivals expressed by an arrival curve. 
pub enum ArrivalModel {
    /// *Periodic* arrivals: task arrivals are evenly spaced in time.
    Periodic(rta::arrival::Periodic),
    /// *Sporadic* arrivals: task arrivals occur unpredictably, but with a minimum inter-arrival time.
    Sporadic(rta::arrival::Sporadic),
    /// Sporadic arrivals bounded by am upper-bounding event *curve*.
    Curve(rta::arrival::ExtrapolatingCurve),
}

impl ArrivalBound for ArrivalModel {
    fn number_arrivals(&self, delta: Duration) -> usize {
        match self {
            Self::Periodic(p) => p.number_arrivals(delta),
            Self::Curve(c) => c.number_arrivals(delta),
            Self::Sporadic(s) => s.number_arrivals(delta),
        }
    }

    fn clone_with_jitter(&self, jitter: Duration) -> Box<dyn ArrivalBound> {
        match self {
            Self::Periodic(p) => p.clone_with_jitter(jitter),
            Self::Curve(c) => c.clone_with_jitter(jitter),
            Self::Sporadic(s) => s.clone_with_jitter(jitter),
        }
    }

    fn steps_iter<'a>(&'a self) -> Box<dyn Iterator<Item = Duration> + 'a> {
        match self {
            Self::Periodic(p) => p.steps_iter(),
            Self::Curve(c) => c.steps_iter(),
            Self::Sporadic(s) => s.steps_iter(),
        }
    }
}

impl TryFrom<&io::Arrival> for ArrivalModel {
    type Error = AppError;

    fn try_from(arr: &io::Arrival) -> Result<Self, AppError> {
        match arr {
            io::Arrival::Period { period: p } => {
                if *p > 0 {
                    Ok(ArrivalModel::Periodic(Periodic::new(Duration::from(*p))))
                } else {
                    Err(AppError::invalid(String::from("periods must be nonzero")))
                }
            }
            io::Arrival::Sporadic { min_inter, jitter } => {
                if *min_inter > 0 {
                    Ok(ArrivalModel::Sporadic(Sporadic::new(
                        Duration::from(*min_inter),
                        Duration::from(*jitter),
                    )))
                } else {
                    Err(AppError::invalid(String::from(
                        "min. inter-arrival times must be nonzero",
                    )))
                }
            }
            io::Arrival::Curve { acp: c } => {
                if c.steps.is_empty() {
                    return Err(AppError::invalid(String::from(
                        "steps sequence must be nonempty",
                    )));
                }
                let mut last = 0;
                for io::Step(x, y) in &c.steps {
                    // skip optional implicit (0, 0) step if explicitly given
                    if *x == 0 && *y == 0 {
                        continue;
                    }
                    if *x >= c.horizon {
                        return Err(AppError::invalid(format!(
                            "step at {} exceeds horizon {}",
                            *x, c.horizon
                        )));
                    }
                    if *y <= last {
                        return Err(AppError::invalid(format!(
                            "steps must be strictly monotonic ({} >= {})",
                            last, *y
                        )));
                    }
                    last = *y;
                }
                let acp = ArrivalCurvePrefix::new(
                    Duration::from(c.horizon),
                    c.steps
                        .iter()
                        // skip optional implicit (0, 0) step if explicitly given
                        .filter(|s| **s != io::Step(0, 0))
                        .map(|io::Step(x, y)| (Duration::from(*x), *y))
                        .collect(),
                );
                Ok(ArrivalModel::Curve(ExtrapolatingCurve::new(Curve::from(
                    acp,
                ))))
            }
        }
    }
}

/// A simple record of a task's parameters. 
pub struct Task {
    /// The task's *worst-case execution time* (WCET, per job). 
    pub cost: wcet::Scalar,
    /// The task's *relative deadline*. 
    pub deadline: Duration,
    /// The task's arrival process. 
    pub arrival_bound: ArrivalModel,
    /// The task's fixed *priority* (if any), where a lower numeric value
    /// corresponds to a lower priority (e.g., like in Linux). 
    pub priority: Option<u32>,
    /// The task's longterm *utilization*.
    pub utilization: f64,
    /// The length of the task's *longest* non-preemptive segment (if
    /// any, not needed by all analyses). 
    pub max_np_segment: Option<Service>,
    /// The length of the task's *last* non-preemptive segment (if any,
    /// not needed by all analyses).     
    pub last_np_segment: Option<Service>,
}

impl TryFrom<&io::Task> for Task {
    type Error = AppError;

    fn try_from(task: &io::Task) -> Result<Self, AppError> {
        if let Some(l) = task.max_nonpreemptive_segment {
            if l > task.wcet {
                return Err(AppError::invalid(format!(
                    "Task {}: max. NP segment exceeds WCET ({} > {})",
                    task.id, l, task.wcet
                )));
            }
        }
        if let Some(l) = task.last_nonpreemptive_segment {
            if l > task.wcet {
                return Err(AppError::invalid(format!(
                    "Task {}: last NP segment exceeds WCET ({} > {})",
                    task.id, l, task.wcet
                )));
            }
        }

        let am = ArrivalModel::try_from(&task.arrivals)?;
        let utilization = match &task.arrivals {
            io::Arrival::Period { period } => task.wcet as f64 / *period as f64,
            io::Arrival::Sporadic {
                min_inter,
                jitter: _,
            } => task.wcet as f64 / *min_inter as f64,
            io::Arrival::Curve { acp } => {
                (am.number_arrivals(Duration::from(acp.horizon)) as u64 * task.wcet) as f64
                    / acp.horizon as f64
            }
        };

        Ok(Task {
            cost: wcet::Scalar::new(Service::from(task.wcet)),
            deadline: Duration::from(task.deadline),
            priority: task.priority,
            arrival_bound: am,
            max_np_segment: task.max_nonpreemptive_segment.map(Service::from),
            last_np_segment: task.last_nonpreemptive_segment.map(Service::from),
            utilization,
        })
    }
}

impl Task {
    /// Obtain the *request-bound function* (RBF) for the task.
    pub fn rbf(&self) -> RBF<&ArrivalModel, &wcet::Scalar> {
        RBF::new(&self.arrival_bound, &self.cost)
    }
}

/// A list of tasks comprising a workload to be jointly executed and analyzed.
pub struct TaskSet {
    /// The underlying list of tasks.
    pub tasks: Vec<Task>,
}

impl TryFrom<&[io::Task]> for TaskSet {
    type Error = AppError;

    fn try_from(tslice: &[io::Task]) -> Result<Self, AppError> {
        let tasks: Result<Vec<Task>, AppError> = tslice.iter().map(Task::try_from).collect();
        Ok(TaskSet { tasks: tasks? })
    }
}

impl TryFrom<&io::ProblemInstance> for TaskSet {
    type Error = AppError;

    fn try_from(pi: &io::ProblemInstance) -> Result<Self, AppError> {
        Self::try_from(&pi.tasks[..])
    }
}

impl TaskSet {

    /// Compute the total utilization of the task set. 
    pub fn total_utilization(&self) -> f64 {
        self.tasks.iter().map(|t| t.utilization).sum()
    }

    /// Ensure that all tasks have a fixed priority, or otherwise return
    /// an error.
    pub fn check_priorities(&self) -> Result<(), AppError> {
        for t in self.tasks.iter() {
            if t.priority.is_none() {
                return Err(AppError::missing("task priority missing".to_string()));
            }
        }
        Ok(())
    }

    /// Ensure that all tasks have the length of their respective *last*
    /// segment specified, or otherwise return an error.
    pub fn check_last_segments(&self) -> Result<(), AppError> {
        for t in self.tasks.iter() {
            if t.last_np_segment.is_none() {
                return Err(AppError::missing(
                    "length of last segment is missing".to_string(),
                ));
            }
        }
        Ok(())
    }

    /// Ensure that all tasks have the length of their respective
    /// *longest* non-preemptive segment specified, or otherwise return an
    /// error.
    pub fn check_max_segments(&self) -> Result<(), AppError> {
        for t in self.tasks.iter() {
            if t.max_np_segment.is_none() {
                return Err(AppError::missing(
                    "max. non-preemptive segment length is missing".to_string(),
                ));
            }
        }
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::io;

    #[test]
    pub fn tasks_from_yaml() {
        let input = "---
task set id: 937723115237613368079497
scheduling policy: EDF
preemption model: FP
task set:

- id: 0
  period: 1173029
  worst-case execution time: 59729
  deadline: 1578636

- id: 1
  period: 5345492
  worst-case execution time: 330336
  deadline: 11631725

- id: 2
  minimum inter-arrival time: 10000
  release jitter: 2500
  worst-case execution time: 2500
  deadline: 18277692

- id: 3
  period: 57242209
  worst-case execution time: 3865381
  deadline: 59575504

- id: 4
  arrival curve: [500, [[0, 0], [1, 1], [15, 2], [100, 3]]]
  worst-case execution time: 10
  deadline: 169499260
...
";
        let pi = io::ProblemInstance::from_yaml(&input).unwrap();
        let ts = TaskSet::try_from(pi.tasks.as_slice()).unwrap();
        let tasks = &ts.tasks;
        assert_eq!(
            tasks[0]
                .arrival_bound
                .number_arrivals(Duration::from(1170000)),
            1
        );
        assert_eq!(
            tasks[0]
                .arrival_bound
                .number_arrivals(Duration::from(1173030)),
            2
        );
        assert_eq!(
            tasks[4].arrival_bound.number_arrivals(Duration::from(30)),
            2
        );
        assert_eq!(
            tasks[4].arrival_bound.number_arrivals(Duration::from(101)),
            3
        );
        assert_eq!(
            tasks[4].arrival_bound.number_arrivals(Duration::from(601)),
            6
        );
        assert_eq!(tasks[0].last_np_segment, None);
        assert_eq!(tasks[1].deadline, Duration::from(11631725));
        assert_eq!(tasks[2].priority, None);
        assert_eq!(tasks[3].max_np_segment, None);
        assert_eq!(tasks[4].cost.wcet, Service::from(10));
        assert_eq!(tasks[2].utilization, 2500 as f64 / 10000 as f64);
        assert_eq!(tasks[4].utilization, 30 as f64 / 500 as f64);
    }
}
