#!/usr/bin/env python3

import yaml
import random
from functools import partial
from pathlib import Path

def ms2ns(x):
    return int(x * 10**6)

def us2ns(x):
    return int(x * 10**3)

# Table 3 in Kramer et al. (WATER 2015)
BOSCH_PERIODS = [
# (period in ms, share among all runnables)
    (1, 0.03),
    (2, 0.02),
    (5, 0.02),
    (10, 0.25),
    (20, 0.25),
    (50, 0.03),
    (100, 0.20),
    (200, 0.01),
    (1000, 0.04),
    (None, 0.15), # angle-synchronous tasks, not relevant here
]
assert sum((p for (_, p) in BOSCH_PERIODS)) == 1.0

def draw_bosch_period(min_period=0, max_period=1000):
    period = None
    while period is None or not (min_period <= period <= max_period):
        x = random.random()
        acc = 0.0
        for (period, prob) in BOSCH_PERIODS:
            acc += prob
            if acc >= x:
                break
    return period

# From Table IV, in us
BOSCH_RUNNABLE_ACETS = {
    1 : (0.34, 5.0, 30.11),
    2 : (0.32, 4.2, 40.69),
    5 : (0.36, 11.04, 83.38),
    10 : (0.21, 10.09, 309.87),
    20 : (0.25, 8.74, 291.42),
    50 : (0.29, 17.56, 92.98),
    100 : (0.21, 10.53, 420.43),
    200 : (0.22, 2.56, 21.95),
    1000 : (0.37, 0.43, 0.46),
}

BOSCH_WCET_FACTORS = {
    1 : (1.30, 29.11),
    2 : (1.54, 19.04),
    5 : (1.13, 18.44),
    10 : (1.06, 30.03),
    20 : (1.06, 15.61),
    50 : (1.13, 7.76),
    100 : (1.02, 8.88),
    200 : (1.03, 4.9),
    1000 : (1.84, 4.75),
}

def draw_bosch_runnable_wcet(period):
    "given a period, return a random WCET in microseconds"
    acet_min, acet_avg, acet_max = BOSCH_RUNNABLE_ACETS[period]
    # bias towards average
    acet = random.triangular(acet_min, acet_max, acet_avg)

    f_min, f_max = BOSCH_WCET_FACTORS[period]
    # bias towards lower end of WCET factor spectrum
    f = random.triangular(f_min, f_max, min(f_min + (f_max - f_min) * 0.1, f_min * 2))

    return acet * f

def draw_runnables(n, min_util=0, max_util=1.0, **kargs):
    ok = 0
    while ok < n:
        p = draw_bosch_period(**kargs)
        c = draw_bosch_runnable_wcet(p)
        p = ms2ns(p)
        c = us2ns(c)
        # filter by utilization bracket
        if min_util <= c / p <= max_util:
            yield (p, c)
            ok += 1

def draw_tasks(n, **kargs):
    # simplification: interpret each runnable as a separate task
    for i, (p, c) in enumerate(sorted(draw_runnables(n, **kargs))):
        yield ("T%02d" % (i + 1), n - i, p, c)

def mk_periodic_task(tid, wcet, deadline, priority, period):
    return {
        "id" : tid,
        "period" : period,
        "worst-case execution time" : wcet,
        "deadline" : deadline,
        "priority" : priority,
    }

def periodic_to_sporadic(tsk, jitter):
    tsk['minimum inter-arrival time'] = tsk['period']
    del tsk['period']
    tsk['release jitter'] = jitter

def mk_sporadic_task(tid, wcet, deadline, priority, min_inter_arrival, jitter):
    return {
        "id" : tid,
        "minimum inter-arrival time" : min_inter_arrival,
        "release jitter" : jitter,
        "worst-case execution time" : wcet,
        "deadline" : deadline,
        "priority" : priority,
    }

def mk_acp_task(tid, wcet, deadline, priority, horizon, steps):
    return {
        "id" : tid,
        "arrival curve" : [horizon, steps],
        "worst-case execution time" : wcet,
        "deadline" : deadline,
        "priority" : priority,
    }

def mk_taskset(id, sched, preempt, tasks):
    return {
        "task set id" : id,
        "scheduling policy" : sched,
        "preemption model" : preempt,
        "task set" : tasks,
    }

def total_util(task_set):
    return sum((t['worst-case execution time']/t['period'] for t in task_set))

def draw_task_sets(start_n = 0, step_size = 1, samples=100, max_n = 100, util_threshold=1.0, tweak_taskset=None, **kargs):
    ref_n = start_n
    min_util = 0
    while (util_threshold is None or min_util <= util_threshold) and (max_n is None or ref_n <= max_n):
        n = max(ref_n, 2)
        task_sets = [[mk_periodic_task(tid, c, p, prio, p) for (tid, prio, p, c) in draw_tasks(n, **kargs)] for _ in range(samples)]
        min_util = min((total_util(ts) for ts in task_sets))
        if tweak_taskset:
            task_sets = [tweak_taskset(ts) for ts in task_sets]
        yield (n, task_sets)
        ref_n += step_size
        while ref_n <= 2: # corner case for step_size = <=2
            ref_n += step_size

def store_task_sets(dest_folder, prefix='', schedulers=['FIFO', 'AFIFO'], preemption_models=['NP'], **kargs):
    dest = Path(dest_folder)
    if dest.exists():
        print("[!!] Destination folder '%s' exists; skipping." %  dest_folder)
        return
    dest.mkdir(parents=True)
    idx = 1
    for (n, task_sets) in draw_task_sets(**kargs):
        for ts in task_sets:
            ts_id = "%s%06d" % (prefix, idx)
            for sched in schedulers:
                for pm in preemption_models:
                    fname = Path(dest, "%s_sched=%s_pm=%s.yaml" % (ts_id, sched, pm))
                    f = fname.open('w')
                    yaml.safe_dump(mk_taskset(ts_id, sched, pm, ts), stream=f, indent=2)
                    f.close()
            idx += 1

def store_unique_task_sets(dest_folder, prefix='', scheduler='FIFO', preemption_model='NP', group_size=1000, **kargs):
    dest = Path(dest_folder)
    if dest.exists():
        print("[!!] Destination folder '%s' exists; skipping." %  dest_folder)
        return
    dest.mkdir(parents=True)
    idx = 0
    for (n, task_sets) in draw_task_sets(**kargs):
        for ts in task_sets:
            if idx % group_size == 0:
                group = Path(dest, "%06d" % idx)
                group.mkdir()
            idx += 1
            ts_id = "%s%06d" % (prefix, idx)
            fname = Path(group, "%s.yaml" % (ts_id))
            f = fname.open('w')
            yaml.safe_dump(mk_taskset(ts_id, scheduler, preemption_model, ts), stream=f, indent=2)
            f.close()


def show_utilizations(**settings):
    for (n, task_sets) in draw_task_sets(**settings):
        utils = [total_util(ts) for ts in task_sets]
        avg_u = sum(utils) / len(utils)
        min_u = min(utils)
        max_u= max(utils)
        feas = sum((1 if u <= 1 else 0 for u in utils))
        print("n=%3d => U: %5.2f %5.2f %5.2f => feas = %6.2f" % (n, min_u, avg_u, max_u, feas / len(utils) * 100))

def assign_jitter(ts, factor=0.75):
    for t in ts:
        jitter = random.randint(0, int(factor * t['period']))
        periodic_to_sporadic(t, jitter)
    return ts

def store_workloads(samples=500):
    store_unique_task_sets(
        './workloads/ex1', prefix='EX1-TS', samples=samples, util_threshold=None,
        max_n=30)
    store_unique_task_sets(
        './workloads/ex2', prefix='EX2-TS', samples=samples, util_threshold=None,
        max_n=80, min_period=20, step_size=2)
    store_unique_task_sets(
        './workloads/ex3', prefix='EX3-TS', samples=samples, util_threshold=None,
        max_n=40, min_period=20, step_size=1, tweak_taskset=assign_jitter)
    store_unique_task_sets(
        './workloads/ex4', prefix='EX4-TS', samples=samples, util_threshold=None,
        max_n=30, tweak_taskset=partial(assign_jitter, factor=1.5))

if __name__ == '__main__':
    store_workloads()
